bl_info = {
	"name": "QnD - Quest and Dialog System",
	"author": "Doc Holiday",
	"version": (1, 1),
	"blender": (2, 79, 0),
	"location": "Node Editor",
	"description": "Adds a new Node System",
	"warning": "Prototype / Study! Use at your own risk!",
	"wiki_url": "http://www.neewo.de/",
	"category": "Game Engine"}

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# QnD_AddOn             2018 by Doc Holiday
# Version: 1.1
# a Part of the QnD Quest and Dialog System
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

import bpy
from bpy.types import NodeTree, Node, NodeSocket
from os import sep as Osep


#                                ╔═════════════════╗
# ▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪ ║ Tree Operations ║
#                                ╚═════════════════╝


class QnD_Tree(NodeTree):
	bl_idname = "QnD_Tree_Type"
	bl_label = "Quest and Dialog System for RPGs"
	bl_icon = "MOD_SOFT"


# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_String_Socket(NodeSocket):		# ◦◦◦◦◦ STRING SOCKET
	bl_idname = "QnD_String_Socket_Type"
	bl_label = "QnD String Socket"

	StringProp = bpy.props.StringProperty()
	showEx = bpy.props.BoolProperty(default = False)

	def draw(self, context, layout, node, text):
		layout.alignment = "EXPAND"
		row = layout.row(align=True)
		row.prop(self, "StringProp", text="")
		
		if len(node.outputs.keys()) > 1:
			if self.showEx:
				TF = "movSock:%s:%s:%s" % (node.name, self.id_data.name, text)
				row.operator("qnd.buttons", text="", icon="TRIA_UP").fnc = "+"+TF
				row.operator("qnd.buttons", text="", icon="TRIA_DOWN").fnc = "-"+TF
				
				TF = "subSock:%s:%s:%s" % (node.name, self.id_data.name, text)
				row.operator("qnd.buttons", text="", icon="PANEL_CLOSE").fnc = TF
		
			row.prop(self, "showEx", text="", icon="GRIP")
				
		


	def draw_color(self, context, node):
		return (0.16, 0.8, 1.0, 0.5)

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_Connector_Green_Socket(NodeSocket):		# ◦◦◦◦◦ CONNECTOR GREEN SOCKET
	bl_idname = "QnD_Connector_Green_Socket_Type"
	bl_label = "QnD Connector Green Socket"

	def draw(self, context, layout, node, text):
		layout.label(text)

	def draw_color(self, context, node):
		return (0.2, 1.0, 0.4, 0.5)

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_Connector_Red_Socket(NodeSocket):			# ◦◦◦◦◦ CONNECTOR RED SOCKET
	bl_idname = "QnD_Connector_Red_Socket_Type"
	bl_label = "QnD Connector Red Socket"

	def draw(self, context, layout, node, text):
		layout.label(text)

	def draw_color(self, context, node):
		return (1.0, 0.2, 0.2, 0.5)

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_Connector_Link_Socket(NodeSocket):			# ◦◦◦◦◦ CONNECTOR LINK SOCKET
	bl_idname = "QnD_Connector_Link_Socket_Type"
	bl_label = "QnD Connector Link Socket"

	def draw(self, context, layout, node, text):
		layout.label("Link IN")
		
		if len(node.outputs.keys()) > 1:
			row = layout.row(align = True)
			TF = "movSock:%s:%s:%s" % (node.name, self.id_data.name, text)
			row.operator("qnd.buttons", text="", icon="TRIA_UP").fnc = "+"+TF
			row.operator("qnd.buttons", text="", icon="TRIA_DOWN").fnc = "-"+TF
			
			TF = "subSock:%s:%s:%s" % (node.name, self.id_data.name, text)
			row.operator("qnd.buttons", text="", icon="PANEL_CLOSE").fnc = TF
			

	def draw_color(self, context, node):
		return (0.8, 0.3, 0.8, 0.5)

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_TreeNode:
	@classmethod
	def poll(cls, ntree):
		return ntree.bl_idname == "QnD_Tree_Type"


#                                ╔═══════╗
# ▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪ ║ Nodes ║
#                                ╚═══════╝


class QnD_Start(Node, QnD_TreeNode):		# ◦◦◦◦◦ START
	bl_idname = "QnD_Start_Node_Type"
	bl_label = "QnD Start Node"
	bl_icon = "FORWARD"

	def init(self, context):
		self.use_custom_color = True
		self.color = (0.6, 0.7, 0.6)
		newSock = self.outputs.new("QnD_Connector_Green_Socket_Type", ">>>")
		newSock.link_limit = 1
		newSock = self.outputs.new("QnD_Connector_Red_Socket_Type", "Cancel")
		newSock.link_limit = 1
		self.width = 100	# default 150

	def draw_label(self):
		return "-- Start --"

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_NPC(Node, QnD_TreeNode):			# ◦◦◦◦◦ NPC
	bl_idname = "QnD_NPC_Node_Type"
	bl_label = "QnD NPC Node"
	bl_icon = "QUESTION"
	
	Spec = bpy.props.StringProperty()
	Snd = bpy.props.StringProperty(subtype="FILE_PATH")

	def init(self, context):
		self.use_custom_color = True
		self.color = (0.6, 0.7, 0.7)
		newSock = self.inputs.new("QnD_Connector_Green_Socket_Type", ">>>")
		newSock.link_limit = 4095
		
		newSock = self.outputs.new("QnD_String_Socket_Type", "")
		newSock.link_limit = 1
		self.width = 200
		
	def draw_buttons(self, context, layout):
		col = layout.column(align=True)
		col.prop(self, "Spec", text="", icon="SCRIPTWIN")
		col.prop(self, "Snd", text="", icon="PLAY_AUDIO")

	def draw_label(self):
		return "-- NPC --"

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_Player(Node, QnD_TreeNode):		# ◦◦◦◦◦ PLAYER
	bl_idname = "QnD_Player_Node_Type"
	bl_label = "QnD Player Node"
	bl_icon = "INFO"
	
	Count = bpy.props.IntProperty()

	def init(self, context):
		self.use_custom_color = True
		self.color = (0.6, 0.6, 0.7)
		self.width = 200
		
		newSock = self.inputs.new("QnD_Connector_Green_Socket_Type", ">>>")
		newSock.link_limit = 4095
		newSock = self.outputs.new("QnD_String_Socket_Type", "0")
		newSock.link_limit = 1

	def draw_buttons(self, context, layout):
		#layout.label("Count: %i" % (self.Count+1))
		row = layout.row(align = True)
		
		TF = "addSock:%s:%s:%i" % (self.name, self.id_data.name, self.Count)
		row.operator("qnd.buttons", text="Choice", icon="ZOOMIN").fnc = TF
		
		TF = "addLink:%s:%s:%i" % (self.name, self.id_data.name, self.Count)
		row.operator("qnd.buttons", text="Link", icon="ZOOMIN").fnc = TF

	def draw_label(self):
		return "-- Player --"

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_Flag(Node, QnD_TreeNode):		# ◦◦◦◦◦ FLAG
	bl_idname = "QnD_Flag_Node_Type"
	bl_label = "QnD Flag Node"
	bl_icon = "MOD_SOFT"
	
	FlagType = bpy.props.EnumProperty(items=[
	("Obj","Object Property","read Object Property"),
	("Sub","Global Sub Dict","read from Global Sub Dictionary"),
	("Glob","Global Dict","read from Global Dictionary"), 
	("State","State","read QnD State")], default="Glob", name="FlagType", description="Flag to read")

	ObjOb = bpy.props.StringProperty()
	ObjProp = bpy.props.StringProperty()
	ObjExp = bpy.props.StringProperty()
	
	Sub = bpy.props.StringProperty()
	SubProp = bpy.props.StringProperty()
	SubExp = bpy.props.StringProperty()
	
	FlagState = bpy.props.IntProperty(default=1, min=0)
	FlagStateExp = bpy.props.StringProperty(default="=")

	def init(self, context):
		self.use_custom_color = True
		self.color = (0.7, 0.6, 0.7)
		self.width = 150
		
		newSock = self.outputs.new("QnD_Connector_Green_Socket_Type", "True:")
		newSock.link_limit = 1
		newSock = self.outputs.new("QnD_Connector_Red_Socket_Type", "False:")
		newSock.link_limit = 1
		
		newSock = self.inputs.new("QnD_Connector_Green_Socket_Type", ">>>")
		newSock.link_limit = 4095

	def draw_buttons(self, context, layout):

		split = layout.split(percentage=.3, align=True)
		split.prop(self, "FlagType", text="", icon_only=True, icon={"Obj":"DOT","Glob":"WORLD","Sub":"URL","State":"TRIA_DOWN_BAR"}[self.FlagType])
		
		if self.FlagType == "Glob" or self.FlagType == "Sub":	# ◦◦◦◦◦ Global Dict / Sub Dir
			if self.FlagType == "Sub":
				split.prop(self, "Sub", text="")
				
				split = layout.split(percentage=0.25, align=True)
				split.label("Var:")
				split.prop(self, "SubProp", text="")
				
			else:
				split.prop(self, "SubProp", text="")
			
			split = layout.split(percentage=0.25, align=True)
			split.label("Exp:")
			split.prop(self, "SubExp", text="")
			
		elif self.FlagType == "Obj":							# ◦◦◦◦◦ Object Property
			split = split.split(percentage=0.9, align=True)
			split.prop_search(self, "ObjOb", bpy.data, "objects", "")
		
			TF = "getActive:%s:%s" % (self.name, self.id_data.name)
			split.operator("qnd.buttons", text="", icon="TRIA_LEFT").fnc = TF
		
			if self.ObjOb:
				layout.prop_search(self, "ObjProp", bpy.data.objects[self.ObjOb].game, "properties", "")

				split = layout.split(percentage=0.25, align=True)
				split.label("Exp:")
				split.prop(self, "ObjExp", text="")
		
		elif self.FlagType == "State":							# ◦◦◦◦◦ State
			split = split.split(percentage=0.4, align=True)
			split.prop(self, "FlagStateExp", text="")
			split.prop(self, "FlagState", text="")



	def draw_label(self):
		return "-- Flag --"

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_Value(Node, QnD_TreeNode):		# ◦◦◦◦◦ VALUE
	bl_idname = "QnD_Value_Node_Type"
	bl_label = "QnD Value Node"
	bl_icon = "MOD_SOFT"
	
	FlagType = bpy.props.EnumProperty(items=[
	("Obj","Object Property","read Object Property"),
	("Sub","Global Sub Dict","read from Global Sub Dictionary"),
	("Glob","Global Dict","read from Global Dictionary"), 
	("State","State","read QnD State")], default="Glob", name="FlagType", description="Flag to read")

	ObjOb = bpy.props.StringProperty()
	ObjProp = bpy.props.StringProperty()
	ObjExp = bpy.props.StringProperty()
	
	Sub = bpy.props.StringProperty()
	SubProp = bpy.props.StringProperty()
	SubExp = bpy.props.StringProperty()
	
	FlagState = bpy.props.IntProperty(default=1, min=0)

	def init(self, context):
		self.use_custom_color = True
		self.color = (0.6, 0.6, 0.5)
		self.width = 150
		
		newSock = self.outputs.new("QnD_Connector_Green_Socket_Type", ">>>")
		newSock.link_limit = 1
		newSock = self.inputs.new("QnD_Connector_Green_Socket_Type", ">>>")
		newSock.link_limit = 4095

	def draw_buttons(self, context, layout):

		split = layout.split(percentage=.3, align=True)
		split.prop(self, "FlagType", text="", icon_only=True, icon={"Obj":"DOT","Glob":"WORLD","Sub":"URL","State":"TRIA_DOWN_BAR"}[self.FlagType])
		
		if self.FlagType == "Glob" or self.FlagType == "Sub":	# ◦◦◦◦◦ Global Dict / Sub Dir
			if self.FlagType == "Sub":
				split.prop(self, "Sub", text="")
				
				split = layout.split(percentage=0.25, align=True)
				split.label("Var:")
				split.prop(self, "SubProp", text="")
				
			else:
				split.prop(self, "SubProp", text="")
			
			split = layout.split(percentage=0.25, align=True)
			split.label("Exp:")
			split.prop(self, "SubExp", text="")
			
		elif self.FlagType == "Obj":							# ◦◦◦◦◦ Object Property
			split = split.split(percentage=0.9, align=True)
			split.prop_search(self, "ObjOb", bpy.data, "objects", "")
		
			TF = "getActive:%s:%s" % (self.name, self.id_data.name)
			split.operator("qnd.buttons", text="", icon="TRIA_LEFT").fnc = TF
		
			if self.ObjOb:
				layout.prop_search(self, "ObjProp", bpy.data.objects[self.ObjOb].game, "properties", "")

				split = layout.split(percentage=0.25, align=True)
				split.label("Exp:")
				split.prop(self, "ObjExp", text="")
		
		elif self.FlagType == "State":							# ◦◦◦◦◦ State
			split.prop(self, "FlagState", text="")

	def draw_label(self):
		return "-- Value --"

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_Level(Node, QnD_TreeNode):		# ◦◦◦◦◦ LEVEL
	bl_idname = "QnD_Level_Node_Type"
	bl_label = "QnD Level Node"
	bl_icon = "INFO"
	
	Count = bpy.props.IntProperty()
	
	FlagType = bpy.props.EnumProperty(items=[
	("Obj","Object Property","read Object Property"),
	("Sub","Global Sub Dict","read from Global Sub Dictionary"),
	("Glob","Global Dict","read from Global Dictionary"), 
	("State","State","read QnD State"),
	("Rand","Random","choose random Level")], default="State", name="FlagType", description="Flag to read")

	ObjOb = bpy.props.StringProperty()
	ObjProp = bpy.props.StringProperty()
	
	Sub = bpy.props.StringProperty()
	SubProp = bpy.props.StringProperty()
	Repeat = bpy.props.BoolProperty(default = False)

	def init(self, context):
		self.use_custom_color = True
		self.color = (0.6, 0.6, 0.6)
		self.width = 140
		
		newSock = self.outputs.new("QnD_Connector_Green_Socket_Type", ">>>  0")
		newSock.link_limit = 1
		newSock = self.inputs.new("QnD_Connector_Green_Socket_Type", ">>>")
		newSock.link_limit = 4095

	def draw_buttons(self, context, layout):
		#split = layout.split(align=True)
		
		split = layout.split(percentage=.3, align=True)
		split.prop(self, "FlagType", text="", icon_only=True, icon={"Obj":"DOT","Glob":"WORLD","Sub":"URL","State":"TRIA_DOWN_BAR","Rand":"QUESTION"}[self.FlagType])

		split = split.split(percentage=(.3,.45)[self.FlagType == "Rand"], align=True)
		TF = "addState:%s:%s:%i" % (self.name, self.id_data.name, self.Count)
		split.operator("qnd.buttons", text="", icon="ZOOMIN").fnc = TF
		
		TF = "subState:%s:%s:%i" % (self.name, self.id_data.name, self.Count)
		split.operator("qnd.buttons", text="", icon="ZOOMOUT").fnc = TF
		
		if self.FlagType != "Rand":
			split.prop(self, "Repeat", text="", icon="FILE_REFRESH")
		
		split=layout.split()
		if self.FlagType == "Glob" or self.FlagType == "Sub":	# ◦◦◦◦◦ Global Dict / Sub Dir
			if self.FlagType == "Sub":
				split.prop(self, "Sub", text="")
				
				split = layout.split(percentage=0.25, align=True)
				split.label("Var:")
				split.prop(self, "SubProp", text="")
				
			else:
				split.prop(self, "SubProp", text="")
			
		elif self.FlagType == "Obj":							# ◦◦◦◦◦ Object Property
			split = split.split(percentage=0.9, align=True)
			split.prop_search(self, "ObjOb", bpy.data, "objects", "")
		
			TF = "getActive:%s:%s" % (self.name, self.id_data.name)
			split.operator("qnd.buttons", text="", icon="TRIA_LEFT").fnc = TF
		
			if self.ObjOb:
				layout.prop_search(self, "ObjProp", bpy.data.objects[self.ObjOb].game, "properties", "")
		
		elif self.FlagType in ("State", "Rand"):							# ◦◦◦◦◦ State
			#print(self.FlagType)
			pass

	def draw_label(self):
		return "-- Level --"

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_XTree(Node, QnD_TreeNode):			# ◦◦◦◦◦ XTREE
	bl_idname = "QnD_XTree_Node_Type"
	bl_label = "QnD XTree Node"
	bl_icon = "CANCEL"
	
	EndState = bpy.props.IntProperty(default=1, min=0)
	Infin = bpy.props.BoolProperty(default = False)
	TX = bpy.props.StringProperty()
	
	

	def init(self, context):
		self.use_custom_color = True
		self.color = (0.5, 0.6, 0.5)
		newSock = self.inputs.new("QnD_Connector_Green_Socket_Type", ">>> X State")
		newSock.link_limit = 4095
		self.width = 150

	def draw_buttons(self, context, layout):
		
		layout.prop_search(self, "TX", bpy.data, "node_groups", "")
		row = layout.row()
		row.prop(self, "Infin", text = "", icon="LOOP_FORWARDS")
		
		if not self.Infin:
			row.prop(self, "EndState", text = "")
		else:
			row.label(text="current")
		
	def draw_label(self):
		return "-- XTree --"
	
# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

class QnD_End(Node, QnD_TreeNode):			# ◦◦◦◦◦ END
	bl_idname = "QnD_End_Node_Type"
	bl_label = "QnD End Node"
	bl_icon = "CANCEL"
	
	EndState = bpy.props.IntProperty(default=1, min=0)
	Infin = bpy.props.BoolProperty(default = False)

	def init(self, context):
		self.use_custom_color = True
		self.color = (0.7, 0.6, 0.6)
		newSock = self.inputs.new("QnD_Connector_Green_Socket_Type", ">>> State")
		newSock.link_limit = 4095
		self.width = 100

	def draw_buttons(self, context, layout):
		row = layout.row()
		row.prop(self, "Infin", text = "", icon="LOOP_FORWARDS")
		
		if not self.Infin:
			row.prop(self, "EndState", text = "")
		else:
			row.label(text="current")
		
	def draw_label(self):
		return "-- End --"


#                                ╔═════════╗
# ▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪ ║ Buttons ║
#                                ╚═════════╝


class QnD_Button(bpy.types.Operator):
	bl_idname = "qnd.buttons"
	bl_label = "QnD Button"
	
	idx = bpy.props.IntProperty()
	fnc = bpy.props.StringProperty()
	
	def execute(self, context):
		
		if self.fnc:
			Inco = self.fnc.split(":")
			do, aNode, aTree = Inco[:3]
			aTree = bpy.data.node_groups[aTree]
			aNode = aTree.nodes[aNode]
			
			if do == "addSock":
				aNode.Count += 1
				newSock = aNode.outputs.new("QnD_String_Socket_Type", str(aNode.Count))
				newSock.link_limit = 1
				aNode.location[1] += 22
			
			elif do == "addLink":
				aNode.Count += 1
				newSock = aNode.outputs.new("QnD_Connector_Link_Socket_Type", str(aNode.Count))
				newSock.link_limit = 1
				aNode.location[1] += 22
				
			elif do == "subSock":
				Nr = Inco[3]
				Socket = aNode.outputs.items()[aNode.outputs.keys().index(Nr)][1]
				aNode.outputs.remove(Socket)
				aNode.location[1] -= 22
			
			elif do[1:] == "movSock":
				Nr = Inco[3]
				ix = aNode.outputs.keys().index(Nr)
				aNode.outputs.move(ix,ix+(1,-1)[do[0] == "+"])
			
			elif do == "getActive":
				aNode.ObjOb = context.scene.objects.active.name
			
			elif do == "addState":
				aNode.Count += 1
				aNode.outputs.new("QnD_Connector_Green_Socket_Type", str(aNode.Count))
				aNode.location[1] += 22
			
			elif do == "subState":
				if aNode.Count > 0:
					aNode.Count -= 1
					Socket = aNode.outputs.items()[-1][1]
					aNode.outputs.remove(Socket)
					aNode.location[1] -= 22
			
		elif self.idx == 1:
			encode("Tree")
			
		elif self.idx == 2:
			encode("all")
		
		elif self.idx == 3:
			LoadMailer()

		self.idx = 0
		self.fnc = ""
		return {'FINISHED'}


#                                ╔═════════════════╗
# ▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪ ║ Node Categories ║
#                                ╚═════════════════╝


import nodeitems_utils
from nodeitems_utils import NodeCategory, NodeItem

class QnDNodeCategory(NodeCategory):
	@classmethod
	def poll(cls, context):
		return context.space_data.tree_type == "QnD_Tree_Type"

# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦

node_categories = [
	QnDNodeCategory("QNDNODES", "Quest & Dialog", items=[
	
		NodeItem("QnD_Start_Node_Type", label="Start"),
		NodeItem("QnD_NPC_Node_Type", label="NPC"),
		NodeItem("QnD_Player_Node_Type", label="Player"),
		NodeItem("QnD_Flag_Node_Type", label="Flag"),
		NodeItem("QnD_Value_Node_Type", label="Value"),
		NodeItem("QnD_Level_Node_Type", label="Level"),
		NodeItem("QnD_XTree_Node_Type", label="XTree"),
		NodeItem("QnD_End_Node_Type", label="End")
		])]

#                                ╔═════════════════╗
# ▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪ ║ File Operations ║
#                                ╚═════════════════╝


def LoadMailer():
	sP = bpy.utils.script_paths("addons%sQnD" % Osep)
	f = "%s%sQnD_Mailer.py" % (sP[0],Osep)
	bpy.ops.text.open(filepath=f, internal=True)

#                                ╔═══════════╗
# ▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪ ║ Sub Menue ║
#                                ╚═══════════╝


class QnD_PT_ToolPanel(bpy.types.Panel):
	bl_category = "Quest & Dialog"
	bl_space_type = "NODE_EDITOR"
	bl_region_type = "TOOLS"
	bl_label = "Tools"
	
	@classmethod
	def poll(cls, context):
		return context.space_data.tree_type == "QnD_Tree_Type"
	
	def draw(self, context):
		
		if "QnD_Mailer.py" not in bpy.data.texts:
			row = self.layout.row()
			row.alert = True
			row.operator("qnd.buttons", text="install Mailer", icon="ERROR").idx=3
		
		box = self.layout.box()
		
		col = box.column()
		col.enabled = (context.active_node != None)
		col.scale_y = 1.5
		col.operator("qnd.buttons", text="encode Tree", icon="FILE_TEXT").idx=1
		col.operator("qnd.buttons", text="encode All", icon="GHOST").idx=2


#                                ╔═══════════════════╗
# ▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪ ║ Encode Operations ║
#                                ╚═══════════════════╝


def encode(w):
	
	# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦ Node Formats: OutLinks; Text; Order
	
	def getOutLinks():
		Result = []
		for lt in N.outputs:
			if lt.links.__len__() > 0:
				Result.append(T.nodes.keys().index(lt.links[0].to_node.name),)
			
			else:
				Result.append("*")
		return tuple(Result)
		#return tuple([T.nodes.keys().index(lt.links[0].to_node.name) for lt in N.outputs if lt.links.__len__() > 0])

	def NF_Start():
		Li = getOutLinks()
		return str(Li)
	
	def NF_NPC():
		Li = getOutLinks()
		Tx = N.outputs[0].StringProp
		Spec = N.Spec
		Snd = N.Snd
		return '%s,"%s","%s",r"%s"' % (str(Li),Tx,Spec,Snd)
	
	def NF_Player():
		Li = getOutLinks()
		Tx = tuple([t.StringProp if t.__class__.__name__ == "QnD_String_Socket" else "__LinkIN__" for t in N.outputs])
		return '%s,%s' % (str(Li),str(Tx))
	
	def NF_Flag(nTy):	# wird auch von Value benutzt
		Li = getOutLinks()
		
		if N.FlagType == "Obj": #Sub,Glob,State
			return '%s,"%s","%s","%s","%s"' % (Li,N.FlagType, N.ObjOb, N.ObjProp, N.ObjExp.replace('"',"'"))
			
		elif N.FlagType == "Sub":
			return '%s,"%s","%s","%s","%s"' %  (Li,N.FlagType, N.Sub, N.SubProp, N.SubExp.replace('"',"'"))
			
		elif N.FlagType == "Glob":
			return '%s,"%s","%s","%s"' %  (Li,N.FlagType, N.SubProp, N.SubExp.replace('"',"'"))
		
		else:	# N.FlagType == "State":
			if nTy == "QnD_Flag":
				return '%s,"%s","%s %i"' %  (Li,N.FlagType, N.FlagStateExp, N.FlagState)
			else:
				return '%s,"%s",%i' %  (Li,N.FlagType, N.FlagState)
	def NF_Level():
		Li = getOutLinks()
		
		if N.FlagType == "Obj": #Sub,Glob,State+Random
			return '%s,"%s","%s","%s",%s' % (Li,N.FlagType, N.ObjOb, N.ObjProp,N.Repeat)
			
		elif N.FlagType == "Sub":
			return '%s,"%s","%s","%s",%s' %  (Li,N.FlagType, N.Sub, N.SubProp,N.Repeat)
			
		elif N.FlagType == "Glob":
			return '%s,"%s","%s",%s' %  (Li,N.FlagType, N.SubProp,N.Repeat)
		
		else:
			return '%s,"%s",%s' %  (Li,N.FlagType,N.Repeat)

	def NF_XTree():
		if N.Infin: Ord = -1
		else: Ord = N.EndState
		return '%s,"%s"' % (Ord,N.TX)

	def NF_End():
		if N.Infin: Ord = -1
		else: Ord = N.EndState
		return '%s' % Ord
		
		
	
	# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦ check File
	fn = "QnD_DataBase.py"
	
	if fn in bpy.data.texts:			# ~~~ File exist
		NewText = bpy.data.texts[fn]
	else:								# ~~~ File new
		bpy.ops.text.new()
		NewText = bpy.data.texts[-1]
		NewText.name = fn
		
		H = '"""\n== Quest & Dialog System '+'='*16+'\nThis is a generated Script. DO NOT TOUCH!\n'+'='*41+'\n"""\n\n'
		NewText.write(H)
	
	if w == "Tree":
		TL = (bpy.context.active_node.id_data,)
	else: # w == "all":
		TL = [t for t in bpy.data.node_groups if t.__class__.__name__ == "QnD_Tree" and t.users > 0]

	for T in TL:
		T.use_fake_user = True
		
		# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦ encoding
		NT = "%s = (\n" % T.name
		for i,N in enumerate(T.nodes):
			# Format: Klasse; Name; ; OutLinks; Text; Order
			Cl = N.__class__.__name__
			Bc = '("%s","%s",%i),' % (Cl,N.name,i)
			BcN = ""
			
			if Cl == "QnD_Start":
				BcN = NF_Start()
			
			elif Cl == "QnD_NPC":
				BcN = NF_NPC()
			
			elif Cl == "QnD_Player":
				BcN = NF_Player()
				
			elif Cl == "QnD_Flag" or Cl == "QnD_Value":
				BcN = NF_Flag(Cl)
			
			elif Cl == "QnD_Level":
				BcN = NF_Level()
			
			elif Cl == "QnD_XTree":
				BcN = NF_XTree()
			
			elif Cl == "QnD_End":
				BcN = NF_End()
			
			NT += "("+ Bc+ BcN+ "),\n"
		
		NT += ")# ||\n\n"
		
		# ◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦◦ finden und ersetzen
		i = NewText.as_string()
		j = i.find("%s = (" % T.name)
		
		if j > 0:							# ~~~ gefunden
			k = i[j:].find(")# ||")
			NT = i.replace(i[j:j+k+7], NT)
			NewText.clear()
		
		NewText.write(NT)

#                                ╔══════════╗
# ▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪▪ ║ Register ║
#                                ╚══════════╝


ClassList =	(
			QnD_Button,
			QnD_Tree,
			QnD_String_Socket,
			QnD_Connector_Green_Socket,
			QnD_Connector_Red_Socket,
			QnD_Connector_Link_Socket,
			QnD_Start,
			QnD_NPC,
			QnD_Player,
			QnD_Flag,
			QnD_Value,
			QnD_Level,
			QnD_XTree,
			QnD_End
			)

def register():
	for cl in ClassList:
		bpy.utils.register_class(cl)

	nodeitems_utils.register_node_categories("QnD_NODES", node_categories)
	bpy.utils.register_module(__name__)


def unregister():
	bpy.utils.unregister_module(__name__)
	nodeitems_utils.unregister_node_categories("QnD_NODES")


if __name__ == "__main__":
	register()
